#!/usr/bin/ruby -w
#
#
# Distributed password cracker 
# 
# Copyright 2008 Raboin Romain - romain.raboin at gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

require 'getoptlong'
require 'webrick'
require 'webrick/https'
require 'digest/md5'
require 'thread'


RANGE 		= 100000000	# default markov range
PRIORITY 	= 1		# default password priority (if 'priority:' is not specified)
PORT		= 8080		# default server port
MKVLVL		= 150		# default markov level
LENMAX		= 6		# default maximum length password
TIMEOUT		= 60*240	# default client timeout in seconds


class	Passwd
	attr_reader	:pass
	attr_reader	:prio
	attr_accessor	:sched
	attr_accessor	:start

	def	initialize(pass, prio, start)
		@pass	= pass	# password string
		@prio	= prio	# password priority
		@sched	= prio	# for scheduling
		@start	= start	# current position in markow range
	end
end


# client information
class	Client
	attr_reader	:num
	attr_accessor	:time
	attr_accessor	:start
	attr_accessor	:end
	attr_accessor	:time
	attr_accessor	:hostname
	attr_accessor	:duration
	attr_accessor	:nbrange
	attr_accessor	:clipasswd

	def	initialize(num, mode)
		@clipasswd	= Array.new	# passwd list to send
		@duration	= 0		# used for client stats
		@hostname	= nil		# client hostname
		@nbrange	= 0		# used for client stats		
		@start		= 0		# markov start
		@time		= Time.now	# 
		@num		= num		# client ID
		@end		= 0		# markov end
		puts "New client id: #{num}"
	end
end


# Manage client:
# - Add new job
# - Check if client timeout
# - Write cracked password
class	ManageClient

	attr_reader	:clientList
	attr_reader	:cracked
	attr_reader	:tfile
	attr_accessor	:file

	def initialize
		@clientList = Hash.new	# list of client
		@delClient  = Array.new	# list of deleted client
		@cracked    = Array.new	# list of cracked account
		@cursched   = 0		# for scheduling password
		@num        = 0		# last id client
		@tfile      = File.open("cracked",  File::WRONLY|File::APPEND|File::CREAT, 0600)
	end

	# password priotity scheduling
	def	findpriority(client)
		begin
			pnext = 0
			psched = $pwdarray[@cursched] 
			if psched.sched == 0
				psched.sched = psched.prio
				pnext = 1
				@cursched = @cursched + 1
				if @cursched == $pwdarray.size
					@cursched = 0
				end
			end
		end while pnext == 1

		client.start = psched.start
		client.end = client.start + $range
		if client.end > $maxpass.to_i
			client.end = $maxpass.to_i
		end

		tmpsched = @cursched
		nbpasswd = 0	
		begin
			psched = $pwdarray[tmpsched]
			psched.sched = psched.sched - 1
			start = psched.start
			prio = psched.prio
			client.clipasswd.push(psched.pass)
			tmpsched = tmpsched + 1
			if tmpsched == $pwdarray.size
				tmpsched = 0			
			end
			nbpasswd = nbpasswd + 1
			tmp = $pwdarray[tmpsched]
			save_start = psched.start		
			psched.start = client.end
		end while nbpasswd < $pwdarray.size and 
			tmp.start == save_start and
			tmp.prio == prio
	end

	# Give new job to the client
	def	newtask(client)
		@clientList.each do |id, cli|
			if Time.now > cli.time + TIMEOUT
				puts "Remove client:#{id} - last request:#{cli.time}"
				@delClient.push cli
				@clientList.delete(id)
				if cli == client
					return "Timeout"
				end
			end
		end
		
		client.clipasswd.clear

		# if a client timeout we must resend its range
		if !@delClient.empty?	
			tmpcli = @delClient.pop;
			client.clipasswd = tmpcli.clipasswd
			client.start = tmpcli.start
			client.end = tmpcli.end
		else
			$mutex.lock
			findpriority(client)
			$mutex.unlock
		end
		nbrange = client.nbrange
		if nbrange > 0
			duration = Time.now - client.time
			client.duration = duration
		end
		nbrange = nbrange + 1
		client.nbrange = nbrange
		client.time = Time.now
	end

	# Client request new job
	def	newjob(cnum, host, method)
		if $pwdarray.size == 0 and @delClient.empty?	# Nothing to do ?
			client = @clientList[cnum.to_i]
			if client != nil and client.start = $maxpass.to_i
				@clientList.delete(cnum.to_i)
			end
			return "Nothing to do\n"		
		end
		if cnum == "0" or cnum == "nil"
			@num = @num + 1
			client = Client.new(@num, method)
			client.hostname = host
			@clientList[@num] = client
			newtask(client)
		else
			client = @clientList[cnum.to_i]
			if client
				newtask(client)
			else
				return "Client id #{cnum} doesn't exist"
			end
		end
		str = "id=#{client.num}\n"
		str = str + "start=#{client.start}\n"
		str = str + "end=#{client.end}\n"
		if method == "markov"
			str = str + "markov=#{$markov}\n"
			str = str + "lenmax=#{$lenmax}\n"		
		end
		str = str + "p=" + client.clipasswd.join("\np=") + "\n"
		$mutex.lock
		$pwdarray.delete_if {|pwd| pwd.start == $maxpass.to_i}	
		$mutex.unlock
		return str
	end
	
	# client find a new password
	def newhash(cnum, hash)
		client = @clientList[cnum.to_i]
		if client
			tab = hash.split(":")
			str = "#{tab[1]}:"
			$mutex.lock
			tfile.write("#{hash}\n")			
			@cracked.push(tab[1])
			$pwdarray.delete_if { |pwd| pwd.pass.index(str, 0) }		
			$mutex.unlock
			puts "New hash from client #{cnum}, user: #{tab[1]}"
		else
			return 	"Client id #{cnum} doesn't exist"
		end
	end
end


class Admin
	
	# Print some stats
	def show
		str = "Server started at: #{$start_time}\n" 
		str += "Maximum length of tested passwords: #{$lenmax}\n"
		str += "Markov level: #{$markov}\n"
		str += "Maximum password: #{$maxpass}\n"
		str += "#{$total_pwd} password(s) loaded for this session.\n"
		str += "#{$mclient.cracked.size} cracked password.\n"
		str += "#{$pwdarray.size} password(s) remaining."

		cli = $mclient.clientList
		if cli.length > 0
			str += "\n\nClient list:\n------------\n"		
		else
			str += "\nClient list is empty\n"
		end
		cli.each do | id, client|
			str += "\nHost name: #{client.hostname}\n"
			str += "Client id: #{client.num}\n"
			time = client.time
			str += "Last request: #{time}\n"		
			duration = client.duration
			if duration > 0
				avg = ($range / duration).to_i
				str += "Average speed #{avg} passwords/second\n"
				finished = time + duration
				endtime = (finished - Time.now).to_i
				if endtime > 0
					str += "Estimated remaing time: #{endtime} second(s)\n"				
				elsif endtime > -3
					str += "Estimated remaing time: 0 second\n"				
				else
					str += "Client Not responding\n"
				end
			end
		end
		#str += "\n\nPassword information:\n--------------------\n\n" #
		#nb = 0
		#$pwdarray.each do | pwd |
		#	nb = nb + 1
		#	percentage = Integer((pwd.start / $maxpass.to_f) * 100)		
		#	str += "Password #{nb}: #{percentage}% done\n"
		#end
		if $mclient.cracked.length > 0
			str += "\n\nCracked account:\n----------------\n\n"
			str = str + "user:" + $mclient.cracked.join("\nuser:") + "\n"	
		end
		return str
	end
end


# HTTP request handle
class PersistAnswers < WEBrick::HTTPServlet::AbstractServlet

	def do_POST(request, response)
	    status, content_type, body = answers_post(request)
	    response.status = status
	    response['Content-Type'] = content_type
	    response.body = body
	end

	def do_GET(request, response)
		status, content_type, body = answers_get(request)
	    response.status = status
	    response['Content-Type'] = content_type
	    response.body = body
	end

	def answers_post(request)
		cpasswd = request.query['passwd']
		cnum = request.query['id']
		if cpasswd != Digest::MD5.hexdigest($password)
			return 403, "text/plain", "Invalid password"
		end
		case request.script_name
		when "/newjob"
			str = $mclient.newjob(cnum, request.query['host'],
				request.query['method']);
			return 200, "text/plain", "#{str}"
		when "/newhash"
			hash = request.query['hash']
			str = $mclient.newhash(cnum, hash)			
			return 200, "text/plain", "OK"
		else
			return 404, "text/plain", "Not found" 	
		end
		return 500, "text/plain", "Server error"
	end
end


# Create passwd array order by priotiry
def	createpwdarray(passtab, start)
	$mutex.lock
	passtab.each do |line|
			tab = line.split(" ",2)
			if tab[0] =~ /priority:/
				prio = tab[0].split(":")
				line = tab[1]
				real_prio = prio[1].to_i
				real_prio = nil if real_prio == 0
			end
			real_prio = real_prio ? real_prio : PRIORITY
			
			# test if password string is already in list			
			if line != nil and $seen[line] == nil
				$seen[line] = 1
				# Check if there's actually a password hash
				if line.split(':')[1] != "*"
					$pwdarray.push(Passwd.new(line, real_prio, start))
				end
			end			
	end
	# Sort password by priotiry
	$pwdarray.sort! { |a, b| b.prio <=> a.prio }
	$total_pwd += $pwdarray.size
	$mutex.unlock
end


def	help
	puts "Usage: #{$0} [options]"
	puts "\t-f --passwdfile\t\tpassword file."
	puts "\t-l --lenmax\t\tthe maximum length of the tested passwords. [#{LENMAX}]"
	puts "\t-m --markov_level\tmaximum strength of passwords that are going to be cracked. [#{MKVLVL}]"
	puts "\t-p --password\t\tserver password for this session."
	puts "\t-P --port\t\tport. [#{PORT}]"
	puts "\t-t --time\t\ttime in seconds before client timeout. [#{TIMEOUT}]"
	puts "\t-r --rangesize\t\tmarkov range size. [#{RANGE}]"
	puts "\t-s --start\t\tindex of fisrt password."
	exit
end


#default value
start 	= 0
port 	= PORT
$time 	= TIMEOUT
$range 	= RANGE
$markov = MKVLVL
$lenmax = LENMAX
$password   = nil
$passwdfile = nil

# mutex for locking $pwdarray
$mutex	= Mutex.new

opts = GetoptLong.new(
	[ '--help', '-h', GetoptLong::NO_ARGUMENT ],
	[ '--passwdfile', '-f', GetoptLong::REQUIRED_ARGUMENT ],
	[ '--lenmax', '-l', GetoptLong::OPTIONAL_ARGUMENT ],
	[ '--markov_level', '-m', GetoptLong::OPTIONAL_ARGUMENT ],
	[ '--password', '-p', GetoptLong::OPTIONAL_ARGUMENT ],
	[ '--port', '-P', GetoptLong::OPTIONAL_ARGUMENT ],
	[ '--range', '-r', GetoptLong::OPTIONAL_ARGUMENT ],	
	[ '--time', '-t', GetoptLong::OPTIONAL_ARGUMENT ],
	[ '--start', '-s', GetoptLong::OPTIONAL_ARGUMENT ]
	)


opts.each do |opt, arg|
	case opt
	when '--help'
		help
	when '--passwdfile'
		$passwdfile = arg
	when '--lenmax'
		$lenmax = arg.to_i
	when '--markov_level'
		$markov = arg.to_i
	when '--password'
		$password = arg
	when '--port'
		port = arg.to_i
	when '--time'
		time = arg.to_i
	when '--range'
		$range = arg.to_i
	when '--start'
		start = arg.to_i
	end
end

if !$password
	puts "Please specify a session password"
	exit
end

if !$passwdfile
	puts "Please specify a password file"
	exit
end



$seen = Hash.new	# 
$pwdarray = Array.new	# Global password list
$start_time = Time.now	# When the server was started
$total_pwd = 0		# Password loaded for this session
passtab = File.read($passwdfile).split("\n")
createpwdarray(passtab, start)


stats = IO::readlines("genmkvpwd/stats.mkv#{$lenmax}.txt")
tab = stats[$markov - 100].split(' ')
$maxpass = tab[1]

$mclient = ManageClient.new


# Webrick ssl configuration
pkey = OpenSSL::PKey::RSA.new(File.open("key.pem").read)
cert = OpenSSL::X509::Certificate.new(File.open("ca.pem").read)
server = WEBrick::HTTPServer.new(
	:Port 		=> port,
  	:SSLEnable	=> true,
  	:SSLVerifyClient => ::OpenSSL::SSL::VERIFY_NONE,
	:SSLCertificate => cert,
	:SSLPrivateKey	=> pkey,
	:SSLCertName	=> [ ["C","FR"], ["O","passwd_cracker.org"], ["CN", "WWW"] ]
	)

# HTTP Authentication configuration
Auth = WEBrick::HTTPAuth::BasicAuth
userdb = Hash.new
realm = "Passwd cracker authentication"
userdb.extend(WEBrick::HTTPAuth::UserDB)
userdb.auth_type = Auth
userdb.set_passwd(realm, "admin", $password)
authenticator = Auth.new({
	:Realm => realm,
	:UserDB => userdb,
	:NonceExpirePeriod => 60,
	:NonceExpireDelta => 5
}) 


server.mount "/newjob",	PersistAnswers
server.mount "/newhash", PersistAnswers
server.mount_proc("/admin") { |req, res|
  	authenticator.authenticate(req, res)
	adm = Admin.new
	str = adm.show
	res.body = "#{str}"
}

trap "INT" do
	$mclient.tfile.close
	server.shutdown	
end

trap "HUP" do
	puts "SIGHUP received"
	passtab = File.read($passwdfile).split("\n")	# reload password file
	createpwdarray(passtab, 0)
end

server.start


